#!/usr/bin/env python3

""" Builds a Docker image locally to run the tests (if it's not already built), then uses it to run
    the massmigration tests.
"""

import os
import subprocess
import sys


IMAGE_NAME = "django-mass-migrations-test"


def abort(msg):
    print(f"🚨 {msg}")
    sys.exit(1)


def check_have_docker():
    try:
        subprocess.check_call(["which", "docker"])
    except subprocess.CalledProcessError:
        abort(
            "It looks like you don't have the 'docker' command installed. "
            "This utility script uses Docker to run the tests."
        )


def project_folder():
    scripts_dir = os.path.dirname(os.path.abspath(__file__))
    return os.path.dirname(scripts_dir)


def dockerfile_path():
    return os.path.join(project_folder(), "testing", "Dockerfile")


def image_exists():
    output = subprocess.check_output(["docker", "images", "-q", IMAGE_NAME])
    return bool(output.strip())


def build_image():
    subprocess.check_call([
        "docker",
        "build",
        "-t",
        IMAGE_NAME,
        "--file",
        dockerfile_path(),
        project_folder(),
    ])


def run_tests():
    subprocess.check_call(
        [
            "docker",
            "run",
            "--rm",
            "--volume",
            f"{project_folder()}/massmigration:/code/massmigration",
            IMAGE_NAME,
            *sys.argv[1:],
        ],
        cwd=project_folder(),
    )


def main():
    check_have_docker()
    if not image_exists():
        build_image()
    run_tests()


if __name__ == '__main__':
    main()



from massmigration.migrations import SimpleMigration


class Migration(SimpleMigration):
    """ YOUR DESCRIPTION HERE. This will appear in the Django admin. """

    # This can be set to make a migration run on a specific backend, rather than the one that's
    # that's specified in the Django settings
    backend: str = None

    # This can be set to specify the list of database aliases on which the migration can be applied.
    # The migration is not forced on a specific DB but rather the `db_alias` for the DB is passed to `operation`
    # allowing to customise what is retrieved by the migration.
    # If None the migration can be applied to all databases.
    allowed_db_aliases: list = None

    dependencies = [{% for dependency in dependencies %}
        ("{{dependency.0}}", "{{dependency.1}}"),{% endfor %}
    ]

    def operation(self, db_alias):

        # PUT YOUR CODE HERE.
        # It should perform the entire migration operation in this single step.
        # This must be complete-able within the memory and time limits of the backend.
        # If you cannot perform the operation within a single operation, then you should use the
        # 'mapper' or 'custom' template instead.

        raise NotImplementedError



from django.urls import path

from . import views


urlpatterns = [
    path("manage/", views.manage_migrations, name="massmigration_manage"),
    path("run/<str:key>/<str:db_alias>/", views.run_migration, name="massmigration_run"),
    path("detail/<str:key>/<str:db_alias>/", views.migration_detail, name="massmigration_detail"),
    path("delete/<str:key>/<str:db_alias>/", views.delete_migration, name="massmigration_delete"),
]



""" A set of utility functions to aid programmatic control of migrations.
    These are really just a collation of functionality from around the app.
"""

# Standard library
from typing import List

# Mass Migration
from . import enforcement
from .exceptions import MigrationAlreadyStarted
from .loader import store
from .migrations import BaseMigration
from .models import MigrationRecord


def get_all_migrations() -> List[BaseMigration]:
    """ Returns a list of all migrations from all installed apps, ordered by app. """
    return store.all


def migration_is_applied(migration: BaseMigration, db_alias: str) -> bool:
    """ Is the given migration applied? """
    return enforcement.migration_is_applied(migration.key, db_alias)


def migration_was_started(migration: BaseMigration, db_alias: str) -> bool:
    """ Has the given migration ever been initiated (even if it hasn't finished or it failed)? """
    return MigrationRecord.objects.using(db_alias).filter(key=migration.key).exists()


def migration_is_in_progress(migration: BaseMigration, db_alias: str) -> bool:
    """ Is the given migration currently running (started but not yet finished and not errored)? """
    return MigrationRecord.objects.using(db_alias).filter(
        key=migration.key, is_applied=False, has_error=False
    ).exists()


def can_start_migration(migration: BaseMigration, db_alias: str) -> bool:
    """ Can the given migration be started? I.e. either it's never been started or all previous
        attempts have errored.
    """
    return not MigrationRecord.objects.using(db_alias).filter(key=migration.key).exclude(
        has_error=True
    ).exists()


def initiate_migration(migration: BaseMigration, db_alias: str) -> bool:
    if migration_is_in_progress(migration):
        raise MigrationAlreadyStarted(f"Migration {migration.key} on db '{db_alias}' is already running.")
    migration.launch(db_alias)



# Third party
from django.test import TestCase

# Mass Migration
from massmigration.loader import (
    is_valid_migration_id,
    is_valid_migration_name,
    migration_id_from_filename,
)

class LoaderTestCase(TestCase):
    """ Tests for the 'loader.py' module. """

    def test_is_valid_migration_id(self):
        """ Test the `is_valid_migration_id` function. """
        for migration_id, expect_valid in [
            ("0023_valid_name", True),
            ("0023_dots_not_allowed.html", False),
            ("no_leading_zeros", False),
        ]:
            self.assertEqual(is_valid_migration_id(migration_id), expect_valid)

    def test_is_valid_migration_name(self):
        """ Test the `is_valid_migration_name` function. """
        for name, expect_valid in [
            ("valid_name_here", True),
            ("hyphens-not-allowed", False),
            ("dots.not.allowed", False),
        ]:
            self.assertEqual(is_valid_migration_name(name), expect_valid)

    def test_migration_id_from_filename(self):
        """ Test the `migration_id_from_filename` function. """
        for filename, expected_id in [
            ("0001_my_migration.py", "0001_my_migration"),
            ("0001_my_migration.html", None),
            ("__init__.py", None),
        ]:
            self.assertEqual(migration_id_from_filename(filename), expected_id)



# Third party
from django.conf import settings
from django.db import router, transaction as django_transaction
from gcloudc.db import transaction as datastore_transaction

# Mass Migration
from massmigration.models import MigrationRecord


# TODO: if we want to add true multi-database support (i.e. allow migrations on two different
# databases within the same project) then we should probably add a `connection_name` attribute to
# the Migration class and pass that into both here and into .using() calls in various places.
# That said, should you be allowed to run the same migration on two different databases?


def get_transaction(db_alias):
    engine = settings.DATABASES[db_alias]["ENGINE"]
    if engine == "gcloudc.db.backends.datastore":
        return datastore_transaction
    return django_transaction



# Standard library
import os
import shutil

# Third party
from django.apps.registry import apps
from django.core.management import execute_from_command_line
from django.test import TestCase

# Mass Migration
from massmigration.constants import MIGRATIONS_FOLDER


class MakeMassMigrationTestCase(TestCase):
    """ Tests for the 'makemassmigration' command. """

    def tearDown(self):
        super().tearDown()
        self.delete_massmigrations_folder()

    def migrations_folder_path(self, app_name="massmigration"):
        """ The expected folder path where the command will put the migration files. """
        app = apps.get_app_config(app_name)
        return os.path.join(app.path, MIGRATIONS_FOLDER)

    def delete_massmigrations_folder(self, app_name="massmigration"):
        """ Delete the folder which the command puts the migration files into. """
        path = self.migrations_folder_path(app_name)
        if os.path.exists(path):
            shutil.rmtree(path)

    def get_migration_file(self, file_name, app_name="massmigration"):
        """ Return the file contents of the given migration file. """
        file_path = os.path.join(self.migrations_folder_path(app_name), file_name)
        with open(file_path) as file:
            return file.read()

    def assert_migration_file_exists(self, file_name, app_name="massmigration"):
        file_path = os.path.join(self.migrations_folder_path(app_name), file_name)
        if not os.path.exists(file_path):
            self.fail(f"Migration file not found at {file_path}")

    def run_command(self, file_name, *args, app_name="massmigration"):
        execute_from_command_line([
            "django-admin",
            "makemassmigration",
            app_name,
            file_name,
            *args,
        ])

    def test_increments_numbers(self):
        self.run_command("first_migration")
        self.assert_migration_file_exists("0001_first_migration.py")
        self.run_command("second_migration")
        self.assert_migration_file_exists("0002_second_migration.py")

    def test_uses_correct_template(self):
        for index, (template, expected_base_class) in enumerate([
            ("simple", "SimpleMigration"),
            ("mapper", "MapperMigration"),
            ("custom", "BaseMigration"),
        ]):
            self.run_command("my_migration", "--template", template)
            file_contents = self.get_migration_file(f"000{index + 1}_my_migration.py")
            self.assertIn(f"\nclass Migration({expected_base_class}):\n", file_contents)



# Standard library
import logging

# Third party
from djangae.processing import (
    datastore_key_ranges,
    firestore_name_key_ranges,
    firestore_scattered_int_key_ranges,
    sequential_int_key_ranges,
    uuid_key_ranges,
)
from djangae.tasks.deferred import defer, defer_iteration_with_finalize
from django.conf import settings
from django.db import models, router
try:
    from gcloudc.db.models.fields.firestore import AutoCharField
except ImportError:
    # This currently only exists in the Firestore branch of gcloudc
    AutoCharField = type("AutoCharField", (), {})

# Massmigration
from massmigration.utils.transaction import get_transaction
from .base import BackendBase

logger = logging.getLogger(__name__)


class DjangaeBackend(BackendBase):
    """ Backend for running operations on Google Cloud Tasks in Djangae projects.
        Works with SQL, Cloud Datastore and Firestore.
    """

    def run_simple(self, migration, db_alias):
        defer(migration.wrapped_operation, db_alias, _queue=self._get_queue_name(migration), _using=db_alias)
        logger.info("Deferred task to run single-task migration %s", migration.key)

    def run_mapper(self, migration, db_alias):
        # Use `defer_iteration_with_finalize` to do the processing with whichever key_ranges_getter
        # is appropriate for the DB.
        queryset = migration.get_queryset(db_alias)
        key_ranges_getter = self._key_ranges_getter(queryset)
        with get_transaction(db_alias).atomic(using=db_alias):
            attempt_uuid = migration.mark_as_started(db_alias)
            defer_iteration_with_finalize(
                queryset,
                self._call_mapper_wrapped_operation,
                self._mark_mapper_as_finished,
                key_ranges_getter=key_ranges_getter,
                migration=migration,
                attempt_uuid=attempt_uuid,
                db_alias=db_alias,
                _queue=self._get_queue_name(migration),
                _transactional=True,
            )
            logger.info("Deferred task to run mapper migration %s", migration.key)

    def _get_queue_name(self, migration):
        """ Get the queue name from settings, or the override on the migration, if set."""
        queue = getattr(migration, "queue_name", None)
        if not queue:
            queue = getattr(settings, "MASSMIGRATION_TASK_QUEUE", None)
        if not queue:
            raise NotImplementedError("Please configure settings.MASSMIGRATION_TASK_QUEUE.")
        return queue

    def _call_mapper_wrapped_operation(self, instance, migration, attempt_uuid, db_alias):
        migration.wrapped_operation(instance, attempt_uuid, db_alias)

    def _mark_mapper_as_finished(self, migration, attempt_uuid, db_alias):
        logger.info("Marking migration %s (attempt %s) as finished.", migration.key, attempt_uuid)
        migration.mark_as_finished(db_alias)

    def _key_ranges_getter(self, queryset):
        # TODO: this could be better at handling the different cases
        connection = router.db_for_write(queryset.model)
        engine = settings.DATABASES[connection]["ENGINE"]
        pk_field = queryset.model._meta.pk
        if isinstance(pk_field, models.UUIDField):
            return uuid_key_ranges
        elif engine == "gcloudc.db.backends.datastore":
            return datastore_key_ranges
        elif engine == "gcloudc.db.backends.firestore":
            if isinstance(pk_field, AutoCharField):
                return firestore_name_key_ranges
            elif isinstance(pk_field, (models.AutoField, models.BigAutoField, models.SmallAutoField)):
                return firestore_scattered_int_key_ranges
        else:  # SQL
            if isinstance(pk_field, models.IntegerField):
                return sequential_int_key_ranges
        # There's also `firestore_scattered_int_key_ranges` which we might want to use in some cases
        raise NotImplementedError(
            f"Key ranges getter function for PKs of type {type(pk_field)} on DB engine '{engine}' "
            "is not yet specified."
        )



# This is the folder name within each installed app in which we look for (and save) migration files
MIGRATIONS_FOLDER = "massmigrations"

DEFAULT_BACKEND = "massmigration.backends.djangae.DjangaeBackend"



from massmigration.migrations import BaseMigration


class Migration(BaseMigration):
    """ YOUR DESCRIPTION HERE. This will appear in the Django admin. """

    # This can be set to make a migration run on a specific backend, rather than the one that's
    # that's specified in the Django settings
    backend: str = None

    # For a custom migration, you must set this, and your backend must have this method.
    backend_method: str = NotImplemented

    # This can be set to specify the list of database aliases on which the migration can be applied.
    # The migration is not forced on a specific DB but rather the `db_alias` for the DB is passed to `launch`
    # allowing to customise what is retrieved by the migration.
    # If None the migration can be applied to all databases.
    allowed_db_aliases: list = None

    dependencies = [{% for dependency in dependencies %}
        ("{{dependency.0}}", "{{dependency.1}}"),{% endfor %}
    ]

    def launch(self, db_alias):
        self.check_dependencies(db_alias)
        self.mark_as_started(db_alias)

        # PUT YOUR CODE HERE.
        # It should defer whatever operation(s) you want to perform to a backend which knows how to
        # perform them.
        # When done, or when an error occurs, the backend must call mark_as_finished(db_alias) or
        # mark_as_errored_(db_alias) on this object as appropriate.

        raise NotImplementedError
